"""Модуль исключений при работе со скважинными данными.
"""


class WellException(BaseException):
    """Базовый класс исключений при работе со скважинными данными"""

    pass


class WellTrajectoryDepthException(WellException):
    """"""

    pass


class WellEmptyCurvesException(WellException):
    """"""

    pass
